from __future__ import annotations

from typing import List
import warnings

from instascrape.core._mappings import _PostMapping, _ProfileMapping
from instascrape.core._static_scraper import _StaticHtmlScraper
from instascrape.scrapers.post import Post


class Profile(_StaticHtmlScraper):
    """
    Scraper for an Instagram profile page

    Methods
    -------
    from_username(shortcode: str) -> Post
        Factory method that returns a Profile object from a shortcode
    """

    _Mapping = _ProfileMapping

    def get_recent_posts(self, amt: int = 12) -> List[Post]:
        """
        Return a list of the profiles recent posts

        Parameters
        ----------
        amt : int
            Amount of recent posts to return

        Returns
        -------
        posts : List[Post]
            List containing the recent 12 posts and their available data
        """
        if amt > 12:
            raise IndexError(f"{amt} is too large, 12 is max available posts")
        posts = []
        post_arr = self.json_dict["entry_data"]["ProfilePage"][0]["graphql"]["user"]["edge_owner_to_timeline_media"][
            "edges"
        ]
        for post in post_arr[:amt]:
            json_dict = post["node"]
            mapping = _PostMapping.post_from_profile_mapping()
            post = Post(json_dict)
            post.load(mapping=mapping)
            posts.append(post)
        return posts

    def _url_from_suburl(self, suburl):
        return f"https://www.instagram.com/{suburl}/"

    @classmethod
    def from_username(self, username):
        warnings.simplefilter('always', DeprecationWarning)
        warnings.warn(
            'This will be deprecated in the near future. You no longer need to use from_username, simply pass username as argument to Profile', DeprecationWarning)
        return Profile(username)



from instascrape.core._mappings import _LoginMapping
from instascrape.core._static_scraper import _StaticHtmlScraper


class LoginAndSignupPage(_StaticHtmlScraper):
    _Mapping = _LoginMapping
    def _construct_url(self):
        pass


from __future__ import annotations

import datetime
from typing import List
import re
import shutil
import pathlib
import warnings

import requests

from instascrape.core._mappings import _PostMapping
from instascrape.core._static_scraper import _StaticHtmlScraper
from instascrape.scrapers.json_tools import parse_json_from_mapping
from instascrape.scrapers.comment import Comment


class Post(_StaticHtmlScraper):
    """
    Scraper for an Instagram post page

    Methods
    -------
    from_shortcode(shortcode: str) -> Post
        Factory method that returns a Post object from a shortcode
    """

    _Mapping = _PostMapping
    SUPPORTED_DOWNLOAD_EXTENSIONS = ['.mp3', '.mp4', '.png', 'jpg']

    def load(self, mapping=None, keys: List[str] = [], exclude: List[str] = []):
        super().load(mapping=mapping, keys=keys, exclude=exclude)

        if mapping is None:
            self.upload_date = datetime.datetime.fromtimestamp(self.timestamp)
            self.tagged_users = self._parse_tagged_users(self.json_dict)
            self.hashtags = self._parse_hashtags(self.caption)

    def download(self, fp: str) -> None:
        """
        Download an image or video from a post to your local machine at the given fpath

        Parameters
        ----------
        fp : str
            Filepath to download the image to
        """
        ext = pathlib.Path(fp).suffix
        if ext not in self.SUPPORTED_DOWNLOAD_EXTENSIONS:
            raise NameError(
                f"{ext} is not a supported file extension. Please use {', '.join(self.SUPPORTED_DOWNLOAD_EXTENSIONS)}"
            )
        url = self.video_url if self.is_video else self.display_url

        data = requests.get(url, stream=True)
        self._download_media(fp, data)

    def get_recent_comments(self):
        list_of_dicts = self.json_dict['entry_data']['PostPage'][0]['graphql']['shortcode_media']['edge_media_to_parent_comment']['edges']
        comments_arr = [Comment(comment_dict) for comment_dict in list_of_dicts]
        return comments_arr

    def to_json(self, fp: str):
        # have to convert to serializable format
        self.upload_date = datetime.datetime.timestamp(self.upload_date)
        super().to_json(fp=fp)
        self.upload_date = datetime.datetime.fromtimestamp(self.upload_date)

    def to_csv(self, fp: str):
        # have to convert to serializable format
        self.upload_date = datetime.datetime.timestamp(self.upload_date)
        super().to_csv(fp=fp)
        self.upload_date = datetime.datetime.fromtimestamp(self.upload_date)

    def _url_from_suburl(self, suburl):
        return f"https://www.instagram.com/p/{suburl}/"

    def _download_media(self, fp: str, data):
        """Write the media to file at given fp from the response"""
        with open(fp, 'wb') as outfile:
            for chunk in data.iter_content(chunk_size=1024):
                if chunk:
                    outfile.write(chunk)
                    outfile.flush()

    def _parse_tagged_users(self, json_dict) -> List[str]:
        """Parse the tagged users from JSON dict containing the tagged users"""
        tagged_arr = json_dict['entry_data']['PostPage'][0]['graphql']['shortcode_media']['edge_media_to_tagged_user']['edges']
        return [node['node']['user']['username'] for node in tagged_arr]

    def _parse_hashtags(self, caption) -> List[str]:
        """Parse the hastags from the post's caption using regex"""
        pattern = r"#(\w+)"
        return re.findall(pattern, caption)

    @classmethod
    def from_shortcode(self, shortcode):
        warnings.simplefilter('always', DeprecationWarning)
        warnings.warn('This will be deprecated in the near future. You no longer need to use from_shortcode, simply pass shortcode as argument to Post', DeprecationWarning)
        return Post(shortcode)



from collections import deque
from collections.abc import MutableMapping
from typing import Any, Dict, List, Union

JSONDict = Dict[str, Any]


class FlatJSONDict(MutableMapping):
    def __init__(self, json_dict):
        """Takes a dictionary with JSON-like data and creates an instance that
        behaves exactly like a dict but with the flattened data"""
        self.json_dict = json_dict

        self.json_tree = JsonTree(self.json_dict)

        self.flat_json = self._flatten_json()

        self.__dict__.update(self.flat_json)

    def _flatten_json(self):
        """
        Primary algorithm that creates the flattened dictionary from a mapped
        tree of JSON data
        """
        flattened_dict = {}
        for leaf_node in self.json_tree.leaf_nodes:
            key_arr = deque([])
            for key in leaf_node.prior_keys[::-1]:
                new_key = self._new_key(key, key_arr)
                if new_key not in flattened_dict:
                    break
            flattened_dict[new_key] = list(leaf_node.json_data.values())[0]
        return flattened_dict

    def _new_key(self, key: str, key_arr: deque) -> str:
        key_arr.appendleft(str(key))
        return "_".join(key_arr)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __delitem__(self, key):
        del self.__dict__[key]

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    # The final two methods aren't required, but nice for demo purposes:

    def __str__(self):
        """returns simple dict representation of the mapping"""
        return str(self.__dict__)

    def __repr__(self):
        """echoes class, id, & reproducible representation in the REPL"""
        return "{}, D({})".format(super(FlatJSONDict, self).__repr__(), self.__dict__)


class JsonTree:
    """Tree of linked lists that map out the JSON data"""

    def __init__(self, json_dict: JSONDict):
        self.json_dict = json_dict
        self.map_tree(self.json_dict)

    def map_tree(self, json_dict):
        """Map the entire JSON tree and get access to leaf nodes"""
        self.leaf_nodes = []
        self.root_node = Node(json_data=json_dict, tree=self)


class Node:
    """
    Representation of one step into a JSON Tree
    """

    def __init__(
        self, json_data: Any, tree: JsonTree, linked_list: deque = deque([]), prior_keys: List[Union[str, int]] = []
    ) -> None:
        self.json_data = json_data
        self.tree = tree
        self.linked_list = linked_list
        self.prior_keys = prior_keys

        self.dtype = type(self.json_data)

        self.nodes = []

        # If the node is a leaf then it has no edges
        if self.is_leaf:
            self.json_data = {prior_keys[-1]: self.json_data}
            self.tree.leaf_nodes.append(self)

        else:
            self.get_edges()

    @property
    def is_leaf(self):
        """
        If the dtype of self.json_data is not a dict or a list then it must be
        a leaf node
        """
        return self.dtype is not list and self.dtype is not dict

    def get_edges(self):
        """
        Get all edges connected to current Node
        """
        iter_arr = zip(range(len(self.json_data)), self.json_data) if self.dtype is list else self.json_data.items()

        for key, value in iter_arr:
            next_linked_list = self.linked_list + deque([self])
            next_key = self.prior_keys + [key]
            node = Node(value, self.tree, next_linked_list, next_key)
            self.nodes.append(node)

    def __repr__(self):
        return str(self.json_data)



from __future__ import annotations

from typing import List
import warnings

from instascrape.core._mappings import _HashtagMapping, _PostMapping
from instascrape.core._static_scraper import _StaticHtmlScraper
from instascrape.scrapers.post import Post


class Hashtag(_StaticHtmlScraper):
    """
    Scraper for an Instagram hashtag page
    """

    _Mapping = _HashtagMapping

    def get_recent_posts(self, amt: int = 71) -> List[Post]:
        """
        Return a list of recent posts to the hasthag

        Parameters
        ----------
        amt : int
            Amount of recent posts to return

        Returns
        -------
        posts : List[Post]
            List containing the recent 12 posts and their available data
        """
        posts = []
        post_arr = self.json_dict["entry_data"]["TagPage"][0]["graphql"]["hashtag"]["edge_hashtag_to_media"]["edges"]
        amount_of_posts = len(post_arr)
        if amt > amount_of_posts:
            amt = amount_of_posts
        for post in post_arr[:amt]:
            json_dict = post["node"]
            mapping = _PostMapping.post_from_hashtag_mapping()
            post = Post(json_dict)
            post.load(mapping=mapping)
            posts.append(post)
        return posts

    def _url_from_suburl(self, suburl):
        return f"https://www.instagram.com/tags/{suburl}/"

    @classmethod
    def from_hashtag(self, hashtag):
        warnings.simplefilter('always', DeprecationWarning)
        warnings.warn(
            'This will be deprecated in the near future. You no longer need to use from_hashtag, simply pass hashtag as argument to Hashtag', DeprecationWarning)
        return Hashtag(hashtag)


from __future__ import annotations

import json
from typing import Any, Dict, Union

import requests
from bs4 import BeautifulSoup

from instascrape.core._json_engine import _JsonEngine

JSONDict = Dict[str, Any]


def parse_json_from_mapping(json_dict, map_dict):
    _json_engine = _JsonEngine(json_dict, map_dict)
    return _json_engine.__dict__


def json_from_html(source: Union[str, BeautifulSoup], as_dict: bool = True) -> Union[JSONDict, str]:
    """
    Return JSON data parsed from Instagram source HTML

    Parameters
    ----------
    source : Union[str, BeautifulSoup]
        Instagram HTML source code to parse the JSON from
    as_dict : bool = True
        Return JSON as dict if True else return JSON as string

    Returns
    -------
    json_data : Union[JSONDict, str]
        Parsed JSON data from the HTML source as either a JSON-like dictionary
        or just the string serialization
    """
    if type(source) is not BeautifulSoup:
        source = BeautifulSoup(source, features="lxml")

    json_script = [str(script) for script in source.find_all("script") if "config" in str(script)][0]
    left_index = json_script.find("{")
    right_index = json_script.rfind("}") + 1
    json_str = json_script[left_index:right_index]

    json_data = json.loads(json_str) if as_dict else json_str
    return json_data


def determine_json_type(json_data: Union[JSONDict, str]) -> str:
    """
    Return the type of Instagram page based on the JSON data parsed from source

    Parameters
    ----------
    json_data: Union[JSONDict, str]
        JSON data that will be checked and parsed to determine what type of page
        the program is looking at (Profile, Post, Hashtag, etc)

    Returns
    -------
    instagram_type : str
        Name of the type of page the program is currently parsing or looking at
    """
    if type(json_data) is not dict:
        json_data = json.loads(json_data)
    instagram_type = list(json_data["entry_data"])[0]
    return instagram_type


def json_from_url(url: str, as_dict: bool = True) -> Union[JSONDict, str]:
    """
    Return JSON data parsed from a provided Instagram URL

    Parameters
    ----------
    url : str
        URL of the page to get the JSON data from
    as_dict : bool = True
        Return JSON as dict if True else return JSON as string

    Returns
    -------
    json_data : Union[JSONDict, str]
        Parsed JSON data from the URL as either a JSON-like dictionary
        or just the string serialization
    """
    source = requests.get(url).text
    json_data = json_from_html(source=source, as_dict=as_dict)
    return json_data



from __future__ import annotations

import datetime
import json
import csv
from abc import ABC, abstractmethod
from typing import Union, Dict, List, Any

import requests
from bs4 import BeautifulSoup

from instascrape.core._json_flattener import FlatJSONDict
from instascrape.scrapers.json_tools import json_from_url, parse_json_from_mapping, json_from_html

# pylint: disable=no-member

JSONDict = Dict[str, Any]

class _StaticHtmlScraper(ABC):
    """
    Base class for all of the scrapers, handles general functionality that all
    scraper objects will have

    Methods
    -------
    load(self, keys: List[str] = [], exclude: List[str] = []) -> None
        Load the static data from the page type
    to_dict(self, metadata: bool = False) -> Dict[str, Any]
        Returns a dictionary of the scraped data
    to_csv(self, fp: str, metadata: bool = False) -> None
        Writes the scraped data to a csv at the given filepath
    to_json(self, fp: str, metadata: bool = False) -> None
        Writes the scraped data to a json file at the given filepath
    """

    # Keys that represent metadata attr that the user doesn't necessarily need
    # to worry about
    _METADATA_KEYS = [
        "json_dict",
        "url",
        "_json_scraper",
        "scrape_timestamp",
        "map_dict",
        "json_data",
        "json_flattener",
        "flat_json_dict",
        "soup",
        "html",
        "source"
    ]

    def __init__(self, source: Union[str, BeautifulSoup, JSONDict]):
        """
        Parameters
        ----------
        url : str
            Full URL to an Instagram page
        name : str
            Optional name for the user to pass
        """
        self.source = source

    def __getitem__(self, key):
        return getattr(self, key)

    def __repr__(self):
        return f"<{type(self).__name__}>"

    def load(self, mapping = None, keys: List[str] = [], exclude: List[str] = []) -> None:
        """
        Scrape data at self.url and parse into attributes

        Parameters
        ----------
        keys : List[str]
            Specify what keys to exclusively load as strings in a list. If no
            keys are specified then the scrape will default load all data points.
        exclude : List[str]
            Specify what keys to exclude from being loaded. If no keys are
            specified, then no data points will be excluded.
        """
        if mapping is None:
            mapping = self._Mapping
        self.json_dict = self._get_json_from_source(self.source)
        self.flat_json_dict = FlatJSONDict(self.json_dict)
        scraped_dict = parse_json_from_mapping(
            json_dict=self.flat_json_dict,
            map_dict=self._Mapping.return_mapping(keys=keys, exclude=exclude),
        )
        self._load_into_namespace(scraped_dict)

    def to_dict(self, metadata: bool = False) -> Dict[str, Any]:
        """
        Return a dictionary containing all of the data that has been scraped

        Parameters
        ----------
        metadata : bool
            Boolean value that determines if metadata specified in self._METADATA_KEYS
            will be included in the dictionary.

        Returns
        -------
        data_dict : Dict[str, str]
            Dictionary containing the scraped data
        """
        data_dict = (
            {key: val for key, val in self.__dict__.items() if key not in self._METADATA_KEYS}
            if not metadata
            else self.__dict__
        )
        return data_dict

    def to_csv(self, fp: str) -> None:
        """
        Write scraped data to .csv at the given filepath

        Parameters
        ----------
        fp : str
            Filepath to write data to
        """
        with open(fp, "w", newline="") as csv_file:
            writer = csv.writer(csv_file)
            for key, value in self.to_dict().items():
                writer.writerow([key, value])

    def to_json(self, fp: str) -> None:
        """
        Write scraped data to .json file at the given filepath

        Parameters
        ----------
        fp : str
            Filepath to write data to
        """
        with open(fp, "w") as outjson:
            json.dump(self.to_dict(), outjson)

    @abstractmethod
    def _url_from_suburl(self, suburl):
        pass

    def _html_from_url(self, url) -> str:
        """Requests page at given URL to get given HTML"""
        response = requests.get(url)
        return response.text

    def _soup_from_html(self, html):
        """Instantiates BeautifulSoup from the given source"""
        return BeautifulSoup(html, features='lxml')

    def json_str_from_soup(self, soup):
        """Instantiates a JSON dictionary from BeautifulSoup"""
        json_script = [str(script) for script in soup.find_all("script") if "config" in str(script)][0]
        left_index = json_script.find("{")
        right_index = json_script.rfind("}") + 1
        json_str = json_script[left_index:right_index]

        return json_str

    def _dict_from_json_str(self, json_str):
        """Load JSON data from a string"""
        return json.loads(json_str)

    def _determine_string_type(self, string_data):
        """Determine's what type of source the string is"""
        string_type_map = [
            ('https://', 'url'),
            ('window._sharedData', 'html'),
            ('{"config"', 'JSON dict str')
        ]
        for substr, str_type in string_type_map:
            if substr in string_data:
                break
        else:
            str_type = 'suburl'
        return str_type

    def _get_json_from_source(self, source):
        """Parses the JSON data out from the source based on what type the source is"""
        source_type = type(source)
        initial_type = True
        if source_type is str:
            source_type = self._determine_string_type(source)
        elif source_type is dict:
            json_dict = source
            return json_dict
        elif source_type is BeautifulSoup:
            source_type = 'soup'

        if source_type == 'suburl':
            if initial_type:
                suburl = self.source
            self.url = self._url_from_suburl(suburl=suburl)
            source_type = 'url'
            initial_type = False

        if source_type == 'url':
            if initial_type:
                self.url = self.source
            self.html = self._html_from_url(url=self.url)
            source_type = 'html'
            initial_type = False

        if source_type == 'html':
            if initial_type:
                self.html = self.source
            self.soup = self._soup_from_html(self.html)
            source_type = 'soup'
            initial_type = False

        if source_type == 'soup':
            if initial_type:
                self.soup = self.source
            json_dict_str = self.json_str_from_soup(self.soup)
            source_type = 'JSON dict str'
            initial_type = False

        if source_type == 'JSON dict str':
            if initial_type:
                json_dict_str = self.source
            json_dict = self._dict_from_json_str(json_dict_str)

        return json_dict

    def _load_into_namespace(self, scraped_dict):
        """Loop through the scraped dictionary and set them as instance attr"""
        for key, val in scraped_dict.items():
            setattr(self, key, val)
        self.scrape_timestamp = datetime.datetime.now()



import datetime

class Comment:
    def __init__(self, comment_dict):
        self.comment_dict = comment_dict['node']

        self._parse_data()

    def __repr__(self):
        return f'<Comment: {self.username}: {self.text}'

    def _parse_data(self):
        self.text = self.comment_dict['text']
        self.created_at = datetime.datetime.fromtimestamp(self.comment_dict['created_at'])
        self.did_report_as_spam = self.comment_dict['did_report_as_spam']
        self.is_verified = self.comment_dict['owner']['is_verified']
        self.profile_pic_url = self.comment_dict['owner']['profile_pic_url']
        self.username = self.comment_dict['owner']['username']
        self.viewer_has_liked = self.comment_dict['viewer_has_liked']
        self.likes = self.comment_dict['edge_liked_by']['count']
        self.is_restricted_pending = self.comment_dict['is_restricted_pending']

        try:
            self.replies = [Comment(comment_dict) for comment_dict in self.comment_dict['edge_threaded_comments']['edges']]
        except KeyError:
            self.replies = []



from __future__ import annotations

from collections import deque
from typing import Any, Dict

JSONDict = Dict[str, Any]


class _JsonEngine:
    """
    Engine for crunching JSON dictionary's using a system of mapping directives
    to access the JSON with a key value for the end user
    """

    DEFAULT_VAL = float("nan")

    def __init__(self, json_data: JSONDict, map_dict: Dict[str, deque]) -> None:
        self.json_data = json_data
        self.map_dict = map_dict

        self._parse_mapping()

    def _parse_mapping(self):
        """
        Loop through each key in map_dict and set value given queue of
        directives that tell the engine how to parse the value from JSONDict
        """
        for key in self.map_dict:
            self._set_value(key, self.json_data, self.map_dict[key])

    def _set_value(self, orig_key: str, container: dict, directive_queue: deque):
        """
        Recursively step through a queue of directives until the queue is
        empty. When there are no more directives left in queue, we have arrived
        at our value and orig_key becomes the user facing attribute and value
        becomes the value of that attribute
        """

        current_key = directive_queue.popleft()
        try:
            value = container[current_key]
        except (KeyError, IndexError, TypeError):
            value = self.DEFAULT_VAL

        if len(directive_queue) == 0:
            setattr(self, orig_key, value)
        else:
            self._set_value(orig_key, value, directive_queue)



from __future__ import annotations

from instascrape.scrapers import *

class InstaScrape:
    """
    Quick and clean interface for tying in much of instascrape's functionality
    in easy and clean syntax
    """
    def load_profile(self, username: str) -> Profile:
        """
        Load, scrape, and return a Profile object that contains all of the
        availably scraped data from a given profile

        Parameters
        ----------
        username : str
            Username of the profile to be scraped

        Returns
        -------
        profile : Profile
            Profile object with the scraped data
        """
        profile = Profile.from_username(username)
        profile.load()
        return profile

    def load_post(self, url: str) -> Post:
        """
        Load, scrape, and return a Post object that contains all of the
        availably scraped data from a given post

        Parameters
        ----------
        url : str
            Full URL of the post to be scraped

        Returns
        -------
        post : Post
            Post object with the scraped data
        """
        post = Post(url)
        post.load()
        return post

    def load_hashtag(self, hashtag: str) -> Hashtag:
        """
        Load, scrape, and return a Hashtag object that contains all of the
        availably scraped data from a given hashtag

        Parameters
        ----------
        hashtag : str
            Name of the hashtag to be scraped

        Returns
        -------
        hashtag : Hashtag
            Hashtag object with the scraped data
        """
        hashtag = Hashtag.from_hashtag(hashtag)
        hashtag.load()
        return hashtag

    def download_post(self, url: str, fp: str) -> None:
        """Download post at the given url to local machine at given fp"""
        post = Post(url)
        post.load()
        post.download(fp)


from instascrape.scrapers.hashtag import *
from instascrape.scrapers.json_tools import *
from instascrape.scrapers.login import *
from instascrape.scrapers.post import *
from instascrape.scrapers.profile import *
from instascrape.scrapers.comment import *